################################################################################
#################################################################################
# ErrorTest_RealData.R
#
# Compute the RMSE of the four variable selection methods (denoted as RKHS, KPS,
# PVS and MH), PLS and the standard L2 regression, using some of the real data 
# sets provided in "generateData.R"
#
# Author: Beatriz Bueno-Larraz
# Date:   22-06-2016
#
# Note: The image of the selected points can be only plotted if all the variable
# selection methods are computed
#################################################################################
#################################################################################
rm(list = ls())

library(fda.usc)
source("selectSeveralMethods.R")
source("generateData.R")
source("Methods/fixedPLS.R") #The original function fregre.pls.cv of fda.usc package has a bug

# DATA AND INITIALIZATION

# Variable selection methods to execute
RKHS = KPS = PVS = MH = T

#type_basis = "pc" #Basis for the L2 estimator
type_basis = "bspline"

# Data sets described in "generateData.R"
data = readTecator()
#data = readSugar()
#data = readFlies() #This one does not appear in the paper
n = nrow(data$x)
Nt = ncol(data$x)
grid = seq(0,1,length=Nt)

# Split the data set into Train (2/3) and Test (1/3)
percTrain = 2/3 #Size of the train set. If it is equal to 0 both sets are equal

if (percTrain == 0) {
  # If the percentaje of the train set is set to 0, train and test sets are equal
  nE = nT = n
  xE = xT = data$x
  yE = yT = data$y
} else {
  # If it is different from 0, the sets are splited
  nE = floor(n*percTrain)
  nT = n - nE
  xE = data$x[1:nE,]
  yE = data$y[1:nE]
  xT = data$x[(nE+1):n,]
  yT = data$y[(nE+1):n]
}


# VARIABLE SELECTION METHODS

# Covariance matrix of the training data set
Sigma = var(xE)

# Compute the selected points of each of the methods 
# (documentation of the function in "selectSeveralMethods.R")
args = list(RKHS=list(Sigma=Sigma), KPS=list(Sigma=Sigma,max=10), MH=list(Sigma=Sigma,max=10))
points = selectSeveralMethods(xE, yE, args, KPS=KPS, PVS=PVS, MH=MH)


# COMPUTATION OF THE ESTIMATED RESPONSES 

# Covariance vector between the trajectories and the responses using the training set
Vcov = array(NA, c(Nt,1))
for (j in 1:Nt) {
  Vcov[j] = cov(xE[,j], yE)
}

# Estimated responses for each of the variable selection methods
yRKHS = yKPS = yPVS = yMH = NA
if (RKHS)  yRKHS = xT[,points$RKHS$ps]%*%solve(Sigma[points$RKHS$ps,points$RKHS$ps])%*%as.matrix(Vcov[points$RKHS$ps], nrow=points$RKHS$p, ncol=1)
if (KPS) yKPS = xT[,points$KPS$ps]%*%solve(Sigma[points$KPS$ps,points$KPS$ps])%*%as.matrix(Vcov[points$KPS$ps], nrow=points$KPS$p, ncol=1)
if (PVS) yPVS = xT[,points$PVS$ps]%*%points$PVS$bs
if (MH) yMH =  xT[,points$MH$ps]%*%solve(Sigma[points$MH$ps,points$MH$ps])%*%as.matrix(Vcov[points$MH$ps], nrow=points$MH$p, ncol=1)

# PLS estimations
pls = fregre.pls.cv_fixed(fdata(xE,argvals=grid), yE, criteria="AIC", lambda=seq(0,5,len=11), P=c(0,0,1))
yPLS = predict(pls$fregre.pls,fdata(xT,argvals=grid))

# Standard L2 estimations
if (type_basis == "pc") {
  l2 = fregre.pc.cv(fdata(xE,argvals=grid),yE,criteria="AIC",kmax=max(floor(Nt/5),11))
  yL2 = predict(l2$fregre.pc,fdata(xT,argvals=grid))
} else {
  l2 = fregre.basis.cv(fdata(xE,argvals=grid),yE)
  yL2 = predict(l2,fdata(xT,argvals=grid))
}


# RESULTS

# Picture of the selected points for each of the variable selection methods
drawPoints = F #Generate the picture or not
saveImage = F #Save the generated picture as a png file or not
name_png="ImageName.png"
if (drawPoints) {
  ylimPlot = c(min(xE),max(xE))
  if (saveImage) png(name_png, width = 800, height = 600)
  par(mfrow=c(2,2))
  
  for (m in 1:4) {
    # Set the title of the plot
    method = (if (m==1) "RKHS" else if (m==2) "KPS" else if (m==3) "PVS" else "MH")
    psTemp = (if (m==1) points$RKHS$ps else if (m==2) points$KPS$ps 
              else if (m==3) points$PVS$ps else points$MH$ps)
    # Draw the curves (25 curves as maximum)
    plot(1:length(xE[1,]), xE[1,], ylim=ylimPlot, type="l", main=method, xlab="", ylab="")
    for (i in 2:min(25,nE)) {
      lines(xE[i,])
    }
    # Draw the selected points
    for (i in 1:length(psTemp)) {
      lines(c(psTemp[i],psTemp[i]), ylimPlot, col="red", lwd=2)
    }
  }
  if (saveImage) dev.off()
}
par(mfrow=c(1,1))

# RMSE of the estimations, if some method is not computed, the error appears as NA
print("PREDICTION ERROR (test set)")
print(paste("RKHS:", format(mean((yT-yRKHS)^2)/(mean(yT^2)),digits=3), 
            ", KPS:", format(mean((yT-yKPS)^2)/(mean(yT^2)),digits=3), 
            ", PVS:", format(mean((yT-yPVS)^2)/(mean(yT^2)),digits=3), 
            ", MH:", format(mean((yT-yMH)^2)/(mean(yT^2)),digits=3), 
            ", PLS:", format(mean((yT-yPLS)^2)/(mean(yT^2)),digits=3),
            ", L2:", format(mean((yT-yL2)^2)/(mean(yT^2)),digits=3))) 
