################################################################################
#################################################################################
# ErrorTest_SimulatedData.R
#
# Compute the mean and the sd of the RMSE of the four variable selection methods 
# (denoted as RKHS, KPS, PVS and MH), PLS and the standard L2 regression, using 
# some of the simulated data sets provided in "generateData.R". If fY1 or fY2 are
# used, also the Haussdorf distance between the estimated points and the real ones 
# and the number of selected points are computed. Boxplots of all these quantities
# are stored. In order to compute the mean and sd of the error, the experiment is
# replicated "mIters" times. These replications are doing using a parallelized loop.
#
# Author: Beatriz Bueno-Larraz
# Date:   28-07-2016
#
# Notes: 
#   - The variable "psreal" defined in line 79 should be changed depending on the
#     regression function that is used (fY1, fY2 or fY3).
#   - The code generates some temporary files that are removed at the end of the
#     execution. If some error occurs, they should be removed manually.
#   - The final results are stored in files using the provided name.
#   - The code of the function HD of the package polydect is included here (under
#     the name dist.hausdorff), in order to avoid loading the complete package.
#################################################################################
#################################################################################
rm(list = ls())

# HD function of the polydect package
dist.hausdorff <-function(X,Y){
  hd=0
  n1<-length(X)
  n2<-length(Y)
  d1<-c()
  d2<-c()
  if(n1*n2==0) hd<- -1
  else{
    stopifnot(is.numeric(X))
    stopifnot(is.numeric(Y))
    for (i in 1:n1)  d1[i]<-min(abs(X[i]-Y))  
    for (i in 1:n2)     d2[i]<-min(abs(Y[i]-X))
    hd=max(max(d1),max(d2))
  }
  return(hd)
}


#################################################################################
#################################################################################

library(fda.usc)
library(foreach)
library(doParallel)
source("selectSeveralMethods.R")
source("generateData.R")
source("Methods/fixedPLS.R") #The original function fregre.pls.cv of fda.usc package has a bug


# DATA AND INITIALIZATION

# Variable selection methods to execute
RKHS = MH = KPS = PVS = T
n_methods = 6 #number of methods to execute

# Basis for the L2 estimator
type_basis = "pc" 
#type_basis = "bspline"

draw = F #If T, the boxplots of the measures are ploted

# Initialization of the parallel setting
n_cores = detectCores()-1 #Always one core should remain free for the operating system
registerDoParallel(n_cores)

name = "BM_fY1" #Name of the simulated data, used to save the results

# Real points used to generate the responses (consult file generateData.R), needed to
# compute the Haussdorf distance between the real points and the estimated ones

psreal = c(20, 40, 90)         #Uncomment if fY1 is used for the responses
#psreal = c(16, 47, 60, 85, 91)  #Uncomment if fY2 is used
#psreal = NULL                  #Uncomment if fY3 is used

mIters = 100  #Number of replications of the experiment

# Parallelized loop over the replications
res = foreach(m=1:mIters) %dopar% {
  # The number of the iterations already executed are written in a temporary file
  write(m, "iters.txt", append=TRUE, ncolumns=1) 
  
  error = list(RKHS=0, KPS=0, PVS=0, MH=0, PLS=0, L2=0)
  nPoints = haus = list(RKHS=0, KPS=0, PVS=0, MH=0)
  
  # Examples of different simulated data (consult file generateData.R)
  
  data = generateBrown(fY1,n=150,SigNoise=0.2, geom=F)  
  #data = generateIntegratedBrown(fY1,n=150,SigNoise=0.2)
  #data = generateOU(fY1,n=150,SigNoise=0.2)
  #data = generateFBM(fY1,n=150,H=0.8,SigNoise=0.2)
  n = nrow(data$x)
  Nt = ncol(data$x)
  grid = seq(0,1,length=Nt)
  
  # Split the data set into Train (2/3) and Test (1/3)
  porcEntr = 2/3 #Size of the train set. If it is equal to 0 both sets are equal
  
  if (porcEntr == 0) {
    # If the percentaje of the train set is set to 0, train and test sets are equal
    nE = nT = n
    xE = xT = data$x
    yE = yT = data$y
  } else {
    # If it is different from 0, the sets are splited
    nE = floor(n*porcEntr)
    nT = n - nE
    xE = data$x[1:nE,]
    yE = data$y[1:nE]
    xT = data$x[(nE+1):n,]
    yT = data$y[(nE+1):n]
  }
  
  
  # VARIABLE SELECTION METHODS
  
  # Covariance matrix of the training data set
  Sigma = var(xE)
  
  # Compute the selected points of each of the methods 
  # (documentation of the function in "selectSeveralMethods.R")
  args = list(RKHS=list(Sigma=Sigma,max=10), KPS=list(Sigma=Sigma,max=10), MH=list(Sigma=Sigma,max=10))
  points = selectSeveralMethods(xE, yE, args, RKHS=RKHS, KPS=KPS, PVS=PVS, MH=MH, mPar=m)
  
  
  # COMPUTATION OF THE HAUSDORFF DISTANCE TO THE REAL POINTS (if fY1 of fY2 are used)
  malla = seq(1/Nt,1,length=Nt)
  if (!is.null(psreal)) {
    if (RKHS) haus$RKHS = dist.hausdorff(malla[points$RKHS$ps],malla[psreal])
    if (KPS) haus$KPS = dist.hausdorff(malla[points$KPS$ps],malla[psreal])
    if (PVS) haus$PVS = dist.hausdorff(malla[points$PVS$ps],malla[psreal])
    if (MH) haus$MH = dist.hausdorff(malla[points$MH$ps],malla[psreal])
  }
  
  
  # COMPUTATION OF THE NUMBER OF SELECTED POINTS (if fY1 of fY2 are used)
  if (!is.null(psreal)) {
    if (RKHS) nPoints$RKHS = points$RKHS$p
    if (KPS) nPoints$KPS =  points$KPS$p
    if (PVS) nPoints$PVS = points$PVS$p  
    if (MH) nPoints$MH = points$MH$p
  }
  
  
  # COMPUTATION OF THE ESTIMATED RESPONSES 
  
  # Covariance vector between the trajectories and the responses using the training set
  Vcov = array(NA, c(Nt,1))
  for (j in 1:Nt)
    Vcov[j] = cov(xE[,j], yE)
  
  # Estimated responses for each of the variable selection methods
  if (RKHS) yRKHS = xT[,points$RKHS$ps]%*%solve(Sigma[points$RKHS$ps,points$RKHS$ps])%*%as.matrix(Vcov[points$RKHS$ps], nrow=points$RKHS$p, ncol=1)
  if (KPS) yKPS = xT[,points$KPS$ps]%*%solve(Sigma[points$KPS$ps,points$KPS$ps])%*%as.matrix(Vcov[points$KPS$ps], nrow=points$KPS$p, ncol=1)
  if (PVS) yPVS = xT[,points$PVS$ps]%*%points$PVS$bs
  if (MH) yMH = xT[,points$MH$ps]%*%solve(Sigma[points$MH$ps,points$MH$ps])%*%as.matrix(Vcov[points$MH$ps], nrow=points$MH$p, ncol=1)
  
  # PLS estimations
  pls = fregre.pls.cv_fixed(fdata(xE,argvals=grid), yE, criteria="AIC", lambda=seq(0,5,len=11), P=c(0,0,1))
  yPLS = predict(pls$fregre.pls,fdata(xT,argvals=grid))
  
  # Standard L2 estimations
  if (type_basis == "pc") {
    l2 = fregre.pc.cv(fdata(xE,argvals=grid),yE,criteria="AIC",kmax=max(floor(Nt/5),11))
    yL2 = predict(l2$fregre.pc,fdata(xT,argvals=grid))
  } else {
    l2 = fregre.basis.cv(fdata(xE,argvals=grid),yE,lambda=seq(0,5,length=11))
    yL2 = predict(l2,fdata(xT,argvals=grid))
  }
  
  # RMSE of the estimations
  if (RKHS) error$RKHS = mean((yT-yRKHS)^2)/mean((yT)^2)
  if (KPS) error$KPS = mean((yT-yKPS)^2)/mean((yT)^2)
  if (PVS) error$PVS = mean((yT-yPVS)^2)/mean((yT)^2)
  if (MH) error$MH = mean((yT-yMH)^2)/mean((yT)^2)
  error$PLS = mean((yT-yPLS)^2)/mean((yT)^2)
  error$L2 = mean((yT-yL2)^2)/mean((yT)^2)
  
  # All the data of each iteration is stored in temporary files, just in case some of the
  # parallelized iterations fails (The number at the end of the file name is used to 
  # distinguish the methods, 1-RKHS 2-KPS 3-PVS 4-MH 5-PLS, 6-L2)
  for (i in 1:4) {
    write(error[[i]], paste0("error_temp", i, ".txt"), append=TRUE)
    write(nPoints[[i]], paste0("nPoints_temp", i, ".txt"), append=TRUE)
    write(haus[[i]], paste0("haus_temp", i, ".txt"), append=TRUE)
  }
  write(error[[5]], paste0("error_temp5.txt"), append=TRUE)
  write(error[[6]], paste0("error_temp6.txt"), append=TRUE)
  
  return(list( error = error, nPoints = nPoints, haus = haus))
}

# End of the parallelization
stopImplicitCluster()


# RESULTS

# All the obtained errors, Haussdorf distances and number of selected points stored in "res" 
# are rearranged in order to print it
error = list(RKHS=rep(0,mIters), KPS=rep(0,mIters), PVS=rep(0,mIters), MH=rep(0,mIters), PLS=rep(0,mIters), L2=rep(0,mIters))
nPoints = haus = list(RKHS=rep(0,mIters), KPS=rep(0,mIters), PVS=rep(0,mIters), MH=rep(0,mIters))
for (m in 1:mIters) {
  if (RKHS) {
    error$RKHS[m] = res[[m]]$error$RKHS
    nPoints$RKHS[m] = res[[m]]$nPoints$RKHS
    haus$RKHS[m] = res[[m]]$haus$RKHS
  } 
  if (KPS) {
    error$KPS[m] = res[[m]]$error$KPS
    nPoints$KPS[m] = res[[m]]$nPoints$KPS
    haus$KPS[m] = res[[m]]$haus$KPS
  }
  if (PVS) {
    error$PVS[m] = res[[m]]$error$PVS
    nPoints$PVS[m] = res[[m]]$nPoints$PVS
    haus$PVS[m] = res[[m]]$haus$PVS
  }
  if (MH) {
    error$MH[m] = res[[m]]$error$MH
    nPoints$MH[m] = res[[m]]$nPoints$MH
    haus$MH[m] = res[[m]]$haus$MH
  }
  error$PLS[m] = res[[m]]$error$PLS
  error$L2[m] = res[[m]]$error$L2
}

# Temporary files are removed
for (i in 1:4) {
  file.remove(paste0("error_temp", i, ".txt"))
  file.remove(paste0("nPoints_temp", i, ".txt"))
  file.remove(paste0("haus_temp", i, ".txt"))
}
file.remove(paste0("error_temp5.txt"))
file.remove(paste0("error_temp6.txt"))
file.remove("iters.txt")

# Information for the box plots for the number of estimated points and the Hausdorff 
# distance is correctly stored, in case fY1 or fY2 are used
# Also the mean and sd of the errors are computed for each method. If some method is
# not computed, -1 appears
boxs = boxsH = boxsE = names = NULL
hausPrint = hausPrintSD = pPrint = pPrintSD = list(RKHS=-1, KPS=-1, PVS=-1, MH=-1)
ePrint = ePrintSD = list(RKHS=-1, KPS=-1, PVS=-1, MH=-1, PLS=-1, L2=-1)

# RKHS method
if (RKHS) {
  if (is.null(boxs))  {
    boxs = nPoints$RKHS
    boxsH = haus$RKHS
    boxsE = error$RKHS
    ePrint$RKHS = mean(error$RKHS); ePrintSD$RKHS = sd(error$RKHS)
    hausPrint$RKHS = mean(haus$RKHS); hausPrintSD$RKHS = sd(haus$RKHS)
    pPrint$RKHS = mean(nPoints$RKHS); pPrintSD$RKHS = sd(nPoints$RKHS)
    names = "RKHS"
  }
} 

# KPS method
if (KPS) {
  if (is.null(boxs)) {
    boxs = nPoints$KPS
    boxsH = haus$KPS
    boxsE = error$KPS
    ePrint$KPS = mean(error$KPS); ePrintSD$KPS = sd(error$KPS)
    hausPrint$KPS = mean(haus$KPS); hausPrintSD$KPS = sd(haus$KPS)
    pPrint$KPS = mean(nPoints$KPS); pPrintSD$KPS = sd(nPoints$KPS)
    names = "KPS"
  } else {
    boxs = cbind(boxs,nPoints$KPS)
    boxsH = cbind(boxsH,haus$KPS)
    boxsE = cbind(boxsE,error$KPS)
    ePrint$KPS = mean(error$KPS); ePrintSD$KPS = sd(error$KPS)
    hausPrint$KPS = mean(haus$KPS); hausPrintSD$KPS = sd(haus$KPS)
    pPrint$KPS = mean(nPoints$KPS); pPrintSD$KPS = sd(nPoints$KPS)
    names = c(names, "KPS")
  }
} 

# PVS method
if (PVS) {
  if (is.null(boxs)) {
    boxs = nPoints$PVS
    boxsH = haus$PVS
    boxsE = error$PVS
    ePrint$PVS = mean(error$PVS); ePrintSD$PVS = sd(error$PVS)
    hausPrint$PVS = mean(haus$PVS); hausPrintSD$PVS = sd(haus$PVS)
    pPrint$PVS = mean(nPoints$PVS); pPrintSD$PVS = sd(nPoints$PVS)
    names = "PVS"
  } else {
    boxs = cbind(boxs,nPoints$PVS)
    boxsH = cbind(boxsH,haus$PVS)
    boxsE = cbind(boxsE,error$PVS)
    ePrint$PVS = mean(error$PVS); ePrintSD$PVS = sd(error$PVS)
    hausPrint$PVS = mean(haus$PVS); hausPrintSD$PVS = sd(haus$PVS)
    pPrint$PVS = mean(nPoints$PVS); pPrintSD$PVS = sd(nPoints$PVS)
    names = c(names, "PVS")
  }
}

# MH method
if (MH) {
  if (is.null(boxs)) {
    boxs = nPoints$MH
    boxsH = haus$MH
    boxsE = error$MH
    ePrint$MH = mean(error$MH); ePrintSD$MH = sd(error$MH)
    hausPrint$MH = mean(haus$MH); hausPrintSD$MH = sd(haus$MH)
    pPrint$MH = mean(nPoints$MH); pPrintSD$MH = sd(nPoints$MH)
    names = "MH"
  } else {
    boxs = cbind(boxs,nPoints$MH)
    boxsH = cbind(boxsH,haus$MH)
    boxsE = cbind(boxsE,error$MH)
    ePrint$MH = mean(error$MH); ePrintSD$MH = sd(error$MH)
    hausPrint$MH = mean(haus$MH); hausPrintSD$MH = sd(haus$MH)
    pPrint$MH = mean(nPoints$MH); pPrintSD$MH = sd(nPoints$MH)
    names = c(names, "MH")
  }
}

# If fY1 or fY2 are used, the box plots for the number of estimated points and the  
# Hausdorff distance are ploted
if (draw) {
  if (!is.null(psreal)) {
    png(paste0(name,"_p.png"), width = 800, height = 600)
    boxplot(boxs, names=names, ylab="p", ylim=c(min(min(boxs,length(psreal))),max(boxs)), main="Number of points p")
    lines(0:(ncol(boxs)+1), rep(length(psreal),(ncol(boxs)+2)),lwd=2,col="red")
    dev.off()
    
    png(paste0(name,"_haus.png"), width = 800, height = 600)
    boxplot(boxsH, names=names, ylab="distance", ylim=c(0,max(boxsH)), main="Hausdorff distance to the real points")
    dev.off()
  }
}

# Plot of the errors
names = c(names, c("PLS","L2"))
boxsE = cbind(boxsE,error$PLS)
boxsE = cbind(boxsE,error$L2)
if (draw) {
  png(paste0(name,"_error.png"), width = 800, height = 600)
  boxplot(boxsE, names=names, ylab="MSE", ylim=c(0,max(boxsE)), main="Mean Squared Error")
  dev.off()
}


# The final data is stored in files and also printed
name = paste0(name,".txt")
if (!is.null(psreal) && any(RKHS,MH,KPS,PVS)) {
  write("n points", name)
  write(t(boxs), name, append=TRUE, ncolumns=100)
  write("haus", name, append=TRUE)
  write(t(boxsH), name, append=TRUE, ncolumns=100)
}
write("error", name, append=TRUE)
write(t(boxsE), name, append=TRUE, ncolumns=100)


ePrint$PLS = mean(error$PLS); ePrintSD$PLS = sd(error$PLS)
ePrint$L2 = mean(error$L2); ePrintSD$L2 = sd(error$L2)


# Print the results in LaTeX format
zerosh = zerose = 5 #compute the scale of the results
for (i in 0:5) {
  if (mean(as.numeric(hausPrint)*(10^i))>1) zerosh = min(zerosh,i)
  if (mean(as.numeric(ePrint)*(10^i))>1) zerose = min(zerose,i)
}

linep = lineh = linee = NULL
for (i in 1:(n_methods-2)) {
  linep = paste0(linep, format(pPrint[[i]],digits=3), " (", format(pPrintSD[[i]],digits=3), ")")
  if (i<(n_methods-2)) linep = paste0(linep, " & ")
  
  lineh = paste0(lineh, format(hausPrint[[i]]*(10^zerosh),digits=3), " (", 
                format(hausPrintSD[[i]]*(10^zerosh),digits=3), ")")
  if (i<(n_methods-2)) lineh = paste0(lineh, " & ")
  
  linee = paste0(linee, format(ePrint[[i]]*(10^zerose),digits=3), " (", 
                format(ePrintSD[[i]]*(10^zerose),digits=3), ") & ")
}
linee = paste0(linee, format(ePrint[[n_methods-1]]*(10^zerose),digits=3), " (", 
               format(ePrintSD[[n_methods-1]]*(10^zerose),digits=3), ") & ") #PLS
linee = paste0(linee, format(ePrint[[n_methods]]*(10^zerose),digits=3), " (", 
               format(ePrintSD[[n_methods]]*(10^zerose),digits=3), ")") #L2

print("METHODS ORDER: RKHS, KPS, PVS, MH, PLS, L2")
print(paste0("Number of selected points (with sd)"))
print(linep)
print(paste0("Haussdorf distance (plus sd). Scale 10^", zerosh))
print(lineh)
print(paste0("RMSE (plus sd). Scale 10^", zerose))
print(linee)


# library(beepr)
# beep()
# beep()
# beep()
# beep()
# beep()
# beep()
# beep()
# beep()
