################################################################################
#################################################################################
# selKPS.R
#
# Contains the function selKPS and the auxiliar function selKPSGivenParams. These 
# functions implement the variable selection procedure described in Kneip et al. 
# (2016)
#
# Author: Beatriz Bueno-Larraz
# Date:   20-05-2016
#
# Note: As mentioned in Section 6.3, only the variable selection procedure of the
# method proposed in Kneip et al. (2016) is implemented.
#################################################################################
#################################################################################


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
# selKPS
#
# By calling the function selKPSGivenParams, execute the variable selection 
# procedure of the mixed regression method proposed in Kneip et al. (2016),
# estimating the parameters if required.
#
# Input
#   x: Data matrix, one row for each curve.
#   y: Vector of response values.
#   Sigma (optional): Covariance matrix of x. If NULL, it is computed.
#   p (optional): Number of points to select. If NULL, it is estimated using the 
#     BIC criterion.
#   pMax (optional): Maximum number of points to select. It is only used if "p"
#     is not provided (15 by default).
#   kd (optional): Parameter k_delta of the method. If NULL, it is estimated
#     using the BIC criterion.
#
# Return
#   Vector containing the set of selected points
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
selKPS <- function(x, y, Sigma=NULL, p=NULL, pMax=15, kd=NULL) {
  Nt = ncol(x) #Grid size
  n = nrow(x) #Sample size
  
  # If some parameter should be estimated, the covariance matrix is needed
  if ((is.null(kd) || is.null(p)) && is.null(Sigma)) {
    Sigma = var(x)
  }
  
  # Covariance vector between the trajectories and the responses
  Vcov = array(NA, c(Nt,1))
  for (j in 1:Nt)
    Vcov[j] = cov(x[,j], y)
  
  # BOTH PARAMETERS SHOULD BE ESTIMATED, BIC criterion is used
  if (is.null(kd) && is.null(p)) {
    kMax = floor((Nt-1)/2) #Maximum value for the k_delta parameter
    BICmin = Inf
    pointsF = pF = kdF = NULL #Temporary variables to store the best values of
    #the parameters during the iterations
    
    # For each value of k_delta
    for (kd in 1:kMax) {
      # Estimated points for each value of p
      for (p in 1:pMax) {
        points = selKPSGivenParams(x,y,n,Nt,kd,p)
        
        # Estimated response, needed to compute the BIC value
        #(we must check that the covariance matrix is invertible for the given points)
        yKPS = tryCatch({
          x[,points]%*%solve(Sigma[points,points])%*%as.matrix(Vcov[points], nrow=p, ncol=1)}, 
          error = function(e) {})
        if(is(yKPS)[1] != "matrix") {break}
        
        BIC = n*mean((y-yKPS)^2) + 2*log(n)
        
        # If the error is reduced in the current iteration, the values are stored
        if (BIC < BICmin) {
          BICmin = BIC
          pointsF = points; pF = p; kdF = kd
        }
      }
    }
    
    # The final values for the points and the parameters are estored
    points = pointsF; p = pF; kd = kdF
    
  } else if (is.null(kd)) {
  # ONLY K_DELTA PARAMETER SHOULD BE ESTIMATED, BIC criterion is used
    kMax = (Nt-1)/2 #Maximum value for the k_delta parameter
    BICmin = Inf
    pointsF = kdF = NULL #Temporary variables to store the best values of
    #the parameters during the iterations
    
    # Estimated points for each value of k_delta and the given (fixed) p
    for (kd in 1:kMax) {
      points = selKPSGivenParams(x,y,n,Nt,kd,p)
      
      # Estimated response, needed to compute the BIC value
      yKPS = x[,points]%*%solve(Sigma[points,points])%*%as.matrix(Vcov[points], nrow=p, ncol=1)
      BIC = n*mean((y-yKPS)^2) + log(n)
      # If the error is reduced in the current iteration, the values are stored
      if ((BIC < BICmin) && (length(points)==p)) {
        BICmin = BIC
        pointsF = points; kdF = kd
      }
    }
    
    # The final values for the points and the parameter k_delta are estored
    points = pointsF; kd = kdF
    
  } else if (is.null(p)) {
    # ONLY P PARAMETER SHOULD BE ESTIMATED, BIC criterion is used
    kMax = (Nt-1)/2 #Maximum value for the k_delta parameter
    BICmin = Inf
    pointsF = pF = NULL #Temporary variables to store the best values of
    #the parameters during the iterations
    
    # Estimated points for each value of p and the given (fixed) k_delta
    for (p in 1:pMax) {
      points = selKPSGivenParams(x,y,n,Nt,kd,p)
      
      # Estimated response, needed to compute the BIC value
      yKPS = x[,points]%*%solve(Sigma[points,points])%*%as.matrix(Vcov[points], nrow=p, ncol=1)
      BIC = n*mean((y-yKPS)^2) + log(n)
      # If the error is reduced in the current iteration, the values are stored
      if (BIC < BICmin) {
        BICmin = BIC
        pointsF = points; pF = p
      }
    }
    
    # The final values for the points and the parameter p are estored
    points = pointsF; p = pF
    
  } else {
  # ALL PARAMETERS ARE GIVEN, nothing to estimate
    points = selKPSGivenParams(x,y,n,Nt,kd,p)
  }
  
  return(list(ps=points,p=p,k=kd))
}



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
# selKPSGivenParams
#
# Execute the variable selection procedure of the mixed regression method proposed 
# in Kneip et al. (2016) for a fixed value of the parameters. The names of the 
# variables follow the notation of the paper.
#
# Input
#   x: Data matrix, one row for each curve.
#   y: Vector of response values.
#   n: Sample size.
#   Nt: Grid size.
#   kd: Parameter k_delta of the method.
#   p: Number of points to select.
#
# Return
#   Vector containing the set of selected points
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
selKPSGivenParams <- function(x,y,n,Nt,kd,p) {
  # INITIALIZATION OF THE PARAMETERS
  delta = kd/Nt #Parameter delta of the paper
  jd = (kd+1):(Nt-kd) #Set J_delta of the paper
  Z = array(NA, c(n,Nt)) #Matrix to store the decorrelated process
  
  # Trajectories of the decorrelated process for each point in J_delta
  for (i in 1:n) {
    for (j in jd) {
      Z[i,j] = x[i,j] - 0.5*(x[i,j-kd] + x[i,j+kd])
    }
  }
  
  # IMPACT POINTS (first step of the algorithm)
  points = array(NA, p)
  for (s in 1:p) {
    # Correlation between the response and the decorrelated process
    corZ = rep(-Inf, Nt)
    for (j in jd) {
      corZ[j] = abs((Z[,j]%*%y)/n)
    }
    
    points[s] = which.max(corZ) #Selected point
    
    # UPDATE THE SET J_delta (second step of the algorithm)
    pos = match(points[s], jd)
    
    # Limits of the interval of size sqrt(delta) around the selected point
    kmin = pos - kd
    kmax = pos + kd

    # If there are still points to be selected, the interval around the selected
    # point is removed from the set J_delta
    if (!any(is.na(jd))) {
      if (kmin < 1) {
        # The interval is [0,kmax]
        jd = jd[kmax:length(jd)]
      } else if (kmax > length(jd)) {
        # The interval is [kmin,1]
        jd = jd[1:kmin]
      } else {
        # The interval is [kmin,kmax]
        jd = c(jd[1:kmin], jd[kmax:length(jd)])
      }
    }
  }
  
  return(sort(unique(points)))
}