#################################################################################
#################################################################################
# selMH.R
#
# Contains the function selMH
#
# Author: Beatriz Bueno-Larraz
# Date:   20-05-2016
#
# Note: MATLAB is required in order to execute this file. If the MATLAB path is
# different from ""/usr/local/MATLAB/R2015a/bin"", it should be changed at the 
# beginning of the function.
# The function creates some temporary files in the working directory, in order
# to comunicate with the MATLAB enviroment, that are removed at the end of the
# execution.
# This file should be located in a folder named "Methods", inside the working
# directory. If not, the "setwd" instruction at the beginning of the function 
# should be changed.
#################################################################################
#################################################################################


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
# selMH
#
# Execute the variable selection method exposed in Berrendero et al. (2016) by
# calling the MATLAB script "maxima_selector.m" provided by José Luis Torrecilla
# (jotorrec@inst.uc3m.es)
#
# Input
#   x: Data matrix, one row for each curve.
#   y: Vector of response values.
#   Sigma (optional): Covariance matrix of x. If NULL, it is computed.
#   p (optional): Number of points to select. If NULL, it is estimated by cross-
#     validation.
#   pMax (optional): Maximum number of points to select. It is only used if "p"
#     is not provided.
#   H (optional): Smoothing parameter that defines the interval to search the
#     maximums.
#   criteria (optional): Criteria to sort the maximums of the function,
#     - 'relevancia': by score (correlation with the class),
#     - 'dominio': width of the interval where the point is a global maximum.
#
# Return
#   List containing the following elements:
#     - ps: Vector containing the set of selected points (sorted)
#     - p: Number of selected points, i.e. length of ps.
#     - h: optimal value of the smoothing parameter
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
selMH <- function(x,y,Sigma=NULL,p=NULL,pMax=15,H=(1:10),criteria="relevancia") {
  library(matlabr)
  library(R.matlab)
  library (energy)
  options(matlab.path = "/usr/local/MATLAB/R2016a/bin") #MATLAB path
  tmpdir = getwd() #Get the current working directory
  setwd(paste0(tmpdir, "/Methods")) #Change the working directory
  
  Nt = ncol(x) #Grid size
  n = nrow(x) #Sample size
  
  # If the covariance matrix is not given, it is computed
  if (is.null(Sigma)) {
    Sigma = var(x)
  }
  
  # Covariance vector between the trajectories and the responses
  Vcov = array(NA, c(Nt,1))
  for (j in 1:Nt) {
    Vcov[j] = cov(x[,j], y)
  }
  
  # Temporary files from which the MATLAB script will read the data
  write(H, "H.txt", append=TRUE, ncolumns=1) #Smoothing parameter
  write(pMax, "dim_red.txt", append=TRUE) #Maximum number of points
  write(criteria, "criterio.txt", append=TRUE) #Criteria to short the maximums
  
  
  # CROSS-VALIDATION
  
  # The optimal value for H is computed by cross-validation
  source(paste0(tmpdir, "/splitCV.R")) #Load the CV code
  kCV = 10 #Number of folds for the CV
  
  # Vector or matrix that stores the errors for each value of the parameters
  error = (if (is.null(p)) array(Inf,c(length(H),pMax)) else rep(0,length(H)))
  
  # For each of the folds in the cross-validation
  for(k in 1:kCV) {
    piecesCV = splitCV(x,y,kCV,k,n) #Current fold
    
    # Distance correlation between the trajectories and the response, using the
    # train data of the current fold
    curve = rep(NA,Nt)
    for (j in 1:Nt) {
      curve[j] = dcor(piecesCV$xE[,j],piecesCV$yE)
    }

    # Store the correlation distances in a temporary file
    write(curve, "curva.txt", append=TRUE, ncolumns=1)
    
    # Run MATLAB script that stores the results in the file "ps.mat"
    run_matlab_script("maxima_selector.m")
    
    # Read the results of the execution from the file
    res = readMat("ps.mat")
    
    
    # SELECTION OF THE POINTS FOR THE CURRENT FOLD
    
    # For each of the values of the smoothing parameter
    for (i in 1:length(H)) {
      # If the number of points should be estimated
      if (is.null(p)) {
        ps = as.vector(res$S[[i]][[1]])
        
        # Compute the error for each number of points. If the returned set of 
        # points is longer than pMax, then it is truncated
        for (j in 1:min(pMax,length(ps))) {
          pstemp = sort(ps[1:j]) #First j selected points
          # Estimated response
          yEst = piecesCV$xV[,pstemp]%*%solve(Sigma[pstemp,pstemp])%*%as.matrix(Vcov[pstemp], nrow=length(pstemp), ncol=1)
          if (error[i,j]==Inf) error[i,j]=0 #If it is the first fold
          error[i,j] = error[i,j] + mean((piecesCV$yV-yEst)^2)/kCV
        }
      } else {
      # If the number of points is given
        pstemp = as.vector(res$S[[i]][[1]]) #All the selected points
        ps = sort(pstemp[1:min(p,length(pstemp))]) #p selected points, or the
        #complete list if the MATLAB script has returned less that p points
        # Estimated response
        yEst = piecesCV$xV[,ps]%*%solve(Sigma[ps,ps])%*%as.matrix(Vcov[ps], nrow=length(ps), ncol=1)
        error[i] = error[i] + mean((piecesCV$yV-yEst)^2)/kCV
      }
    }
    
    # Remove the temporary files created in this iteration of the loop
    file.remove("ps.mat")
    file.remove("curva.txt")
  }
  
  
  # POINTS USING THE OPTIMAL PARAMETERS
  
  # Find the parameters with less error
  if (is.null(p)) {
    # If the number of points and the smoothing parameter are estimated
    mins = which(error==min(error), arr.ind = TRUE)
    h = H[mins[1]]
    p = mins[2]
  } else {
    # If only the smoothing parameter is estimated
    h = H[which.min(error)]
  }
  
  # Distance correlation between the trajectories and the response, using the
  # complete train set
  curve = rep(NA,Nt)
  for (j in 1:Nt) {
    curve[j] = dcor(x[,j],y)
  }
  
  # Temporary files from which the MATLAB script will read the data
  write(curve, "curva.txt", append=TRUE, ncolumns=1) #Distance correlations
  file.remove("H.txt") #Previous file used for cross-validation
  write(h, "H.txt") #Optimal value for the smoothing parameter
  
  # Run MATLAB script and read the results
  run_matlab_script("maxima_selector.m")
  res = readMat("ps.mat") 
  pstemp = as.vector(res$S[[1]][[1]]) #All the selected points
  ps = sort(pstemp[1:min(p,length(pstemp))]) #p selected points, or the
  #complete list if the MATLAB script has returned less that p points
  
  # Remove all the temporary files
  file.remove("H.txt")
  file.remove("dim_red.txt")
  file.remove("criterio.txt")
  file.remove("ps.mat")
  file.remove("curva.txt")
  
  # Change the working directory to the original one
  setwd(tmpdir)
  
  return(list(ps=ps,p=length(ps),h=h))
}



