#################################################################################
#################################################################################
# selPVS.R
#
# Contains the function selPVS and the auxiliar function selPVSGivenParams. These 
# functions implement the variable selection procedure described in Aneiros and
# Vieu (2014)
#
# Author: Beatriz Bueno-Larraz
# Date:   20-05-2016
#
# Note: This method does not allow to select the final number of selected points
#################################################################################
#################################################################################


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
# selPVS
#
# By calling the function selPVSGivenParams, execute the variable selection 
# procedure of the mixed regression method proposed in Aneiros and Vieu (2014),
# estimating the parameters if required.
#
# Input
#   x: Data matrix, one row for each curve.
#   y: Vector of response values.
#   wn (optional): Number of variables to select in the first step of the method
#     (w_n of the paper). If NULL, it is estimated by cross-validation.
#   wnMax (optional): Maximum number of variables to select in the first step of 
#     the algorithm. If NULL, it is setted as half of the grid size.
#
# Return
#   List containing the following elements:
#     - ps: Vector containing the set of selected points (sorted)
#     - p: Final number of selected points, i.e. length of ps.
#     - wn: Optimal value for the number of selected variables in the first step.
#     - bs: Weights associated to the selected points.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
selPVS <- function(x,y,wn=NULL,wnMax=NULL) {
  library(glmnet)
  n = nrow(x) #Sample size
  Nt = ncol(x) #Grid size
  
  # If the maximum number of points is not given, it is setted to the half of the grid points
  if (is.null(wnMax)) {
    wnMax = Nt/2
  }
  
  # If the maximum number of points for the first step of the algorithm is not given, it is
  # estimated by cross-validation
  if (is.null(wn)) {
    wnGrid = 2:wnMax #Possible values for the number of variables to select
    minGCV = wnF = Inf #Minimum value of the error until the current iteration
    resPVS = NULL #Temporary values for the selected points and parameters
    
    # For each value of w_n
    for (wn in wnGrid) {
      # Estimated points for each value of w_n
      resPVStemp = selPVSGivenParams(x,y,Nt,wn)
      
      # Error of the estimated response
      H = x%*%solve(t(x)%*%x+wn*diag(Nt))%*%t(x)
      error = mean((y-x[,resPVStemp$points]%*%as.matrix(resPVStemp$coefs,nrow=n,ncol=1))^2/sum(diag(n)-H)^2)
      
      # If the error is reduced in the current iteration, the values are stored
      if (error < minGCV) {
        minGCV = error
        resPVS = resPVStemp
        wnF = wn
      }
    }
  } else {
  # If w_n given, nothing to estimate
    resPVS = selPVSGivenParams(x,y,Nt,wn)
  }
  
  return(list(ps=resPVS$points, p=length(resPVS$points), wn=wnF, bs=resPVS$coefs))
}



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
# selPVSGivenParams
#
# Execute the variable selection procedure of the mixed regression method proposed 
# in Kneip et al. (2016) for a fixed value of the parameters. The names of the 
# variables follow the notation of the paper.
#
# Input
#   x: Data matrix, one row for each curve.
#   y: Vector of response values.
#   Nt: Grid size.
#   wn: Number of variables to select in the first step of the method (w_n of the paper).
#
# Return
#   List containing the following elements:
#     - points: Vector containing the set of selected points.
#     - coefs: Weights associated to the selected points.
#     - error: Optimal error given by the LASSO method
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
selPVSGivenParams <- function(x,y,Nt,wn) {
  qn = floor(Nt/wn) #q_n parameter of the method
  n = nrow(x) #Sample size
  n1 = floor(n/2) #n_1 parameter of the method: Size of the sample for the first step
  n2 = n - n1 #n_2 parameter of the method: Size of the sample for the second step
  
  # The samples are splitted for the two steps of the method
  x1 = x[1:n1,]
  y1 = y[1:n1]
  x2 = x[(n1+1):n,]
  y2 = y[(n1+1):n]
  
  
  # FIRST STEP
  
  # Equispaced subset of variables Q_n
  Qn = NULL
  for (k in 1:wn) {
    Qn = cbind(Qn, x1[,((2*k-1)*qn/2)])
  }
  
  # Parameters for cross-validation used in LASSO with ML penalization
  kCV = 10 #Number of folds for the CV
  nCV = floor(n1/kCV) #Sample size of each fold
  foldid = rep(kCV,n1) #Initialization of the parameter for cv.glmnet function
  for (k in 1:(kCV-1)) {
    foldid[((k-1)*nCV+1):(k*nCV)] = k
  }
  
  # Generalized linear model via penalized maximum likelihood
  reg1 = cv.glmnet(Qn,y1,nfolds=kCV,foldid=foldid,standardize=F) 
  
  # Coefficients of the points for the optimal value of lambda
  coefs1 = as.vector(coef(reg1, s=reg1$lambda.min))[2:(wn+1)] 
  
  # Points such that their coefficients are different from 0
  SQn = setdiff(1:wn, which(coefs1 %in% 0)) 
  
  # If all the coefficients are zero, then all the variables are used in the next step
  if (all(coefs1==0)) {
    SQn = 1:wn
  }
  

  # SECOND STEP
  # All the variables that are close from those selected in step 1 are stored in R_n
  Rn = NULL
  pos2 = NULL
  # For each of the selected variables
  for (k in SQn) {
    # The inteval around them is included
    for (j in ((k-1)*qn+1):(k*qn)) {
      Rn = cbind(Rn, x2[,j])
      pos2 = c(pos2, j)
    }
  }
  
  # Parameters for cross-validation used in LASSO with ML penalization
  nCV = floor(n2/kCV) #Sample size of each fold
  foldid = rep(kCV,n2) #Initialization of the parameter for cv.glmnet function
  for (k in 1:(kCV-1)) {
    foldid[((k-1)*nCV+1):(k*nCV)] = k
  }
  
  # Generalized linear model via penalized maximum likelihood
  reg2 = cv.glmnet(Rn,y2,nfolds=kCV,foldid=foldid,standardize=F) 
  
  # Coefficients of the points for the optimal value of lambda
  coefs2 = as.vector(coef(reg2, s=reg2$lambda.min))[2:(ncol(Rn)+1)]
  
  # Points such that their coefficients are different from 0
  dn = setdiff(1:ncol(Rn), which(coefs2 %in% 0))
  
  # Final values for the points and the weights
  points = pos2[dn]
  coefs = coefs2[dn]
  
  return(list(points=points, coefs=coefs, error=reg2$cvm[reg2$lambda.min]))
}




