################################################################################
#################################################################################
# selRKHS.R
#
# Contains the function selRKHS
#
# Author: Beatriz Bueno-Larraz
# Date:   31-01-2017
#
# Note: It is not the file used for the simulations of the paper. In this version
# the loops have been removed in order to improve the efficiency. The selected 
# points are the same but execution times could change.
#################################################################################
#################################################################################


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
# selRKHS
#
# Execute the variable selection method exposed in the paper
#
# Input
#   x: Data matrix, one row for each curve.
#   y: Vector of response values.
#   Sigma (optional): Covariance matrix of x. If NULL, it is computed.
#   p (optional): Number of points to select. If NULL, it is estimated using the 
#     method described in Section 4.1 of the paper.
#   pMax (optional): Maximum number of points to select. It is only used if "p"
#     is not provided. Is should be large enough number to ensure that the
#     values Q_n^max(p+1)-Q_n^max(p) converge to zero (15 by default).
#
# Return
#   List containing the following elements:
#     - ps: Vector containing the set of selected points (sorted)
#     - p: Number of selected points, i.e. length of ps.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
selRKHS <- function(x, y, Sigma=NULL, p=NULL, pMax=10) {
  Nt = ncol(x) #Grid size

  # If the covariance matrix is not given, it is computed
  if (is.null(Sigma)) {
    Sigma = var(x)
  }
  
  # Covariance vector between the trajectories and the responses
  Vcov = array(NA, c(Nt,1))
  for (j in 1:Nt) {
    Vcov[j] = cov(x[,j], y)
  }
  
  # If the number of points to estimate is not given, it is setted to the
  # maximum and later on it is estimated
  estimateP = F
  if(is.null(p)){
    estimateP = T
    p = pMax
  }
  
  # First point: using the greedy method described in Section 3.3
  Quot = Vcov^2/diag(Sigma) #Quotient of Equation (22)
  MaxVal = rep(NA, p) #Maximum value of Quot for each p (Q_n^max(p+1)-Q_n^max(p))
  MaxVal[1] = max(Quot)
  pos = which(Quot==MaxVal[1])
  pointsOrig = points = pos #Set of points T_p: "points" is the sorted version
  
  # Rest of the points
  for (j in 2:p) {
    news = setdiff(1:Nt,points) #Points that are still not selected
    inverse = solve(Sigma[points,points]) #Matrix Sigma_{T_p}^{-1}
    
    # Numerator of quotient in Equation (22)
    numerator = (Sigma[news,points]%*%inverse%*%Vcov[points] - Vcov[news])^2
    
    # Denominator of quotient in Equation (22)
    denominator = diag(Sigma)[news] - diag(Sigma[news,points]%*%inverse%*%Sigma[points,news])
    
    Quot = as.numeric(numerator/denominator)
    MaxVal[j] = max(Quot)
    pos = news[which(Quot==MaxVal[j])] #The points that maximized the quotient is selected
    pointsOrig = c(pointsOrig,pos)
    points = sort(c(points,pos))
  }

  # If required, the number of points is estimated (Section 4.1)
  if (estimateP) {
    vals = log(MaxVal[2:p]) #Always one point should be selected as minimum
    # If some value in MaxVal is very small, the logarithm could return NA
    if (any(is.na(vals))) {
      # If it is the case, it is setted to the minimal value
      vals[is.na(vals)] = min(vals[!is.na(vals)])
    }
    
    # The centers for kMeans are fixed to the minimum and maximum values of the set
    minVals = min(vals); maxVals = max(vals) 
    res = kmeans(vals, c(minVals,(minVals+maxVals)/2)) #kMeans method
    
    c1 = res$cluster[1] #Cluster of the first point
    p = tail(which(res$cluster==c1),1) + 1 #p is equal to the last point that belongs to c1
    points = sort(pointsOrig[1:min(p,pMax)]) #The complete set of points is truncated
  }
  
  return(list(ps=points,p=p))
}
