################################################################################
#################################################################################
# generateData.R
#
# Functions that generate the different data sets, both simulated and real, used 
# for the experiments. The return of the functions is always a list of two elements:
#   - x: Data matrix, one row for each curve
#   - y: Vector of responses (real numbers)
#
# Author: Beatriz Bueno-Larraz
# Date:   20-05-2016
# Modified: 07-03-2018
#
# Note: The code of some functions of other existing packages are included here
# (pracma, Sim.DiffProc and dvfBm), in order to avoid loading the whole packages
#################################################################################
#################################################################################


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
# regression functions
#
# Each of the regression functions used to generate the simulated data, described
# in section 6.1 of the paper.
#   - fY1: Sparse beta function, generates Y = 2*X(0.2) - 5*X(0.4) + X(0.9) + E
#   - fY2: Sparse beta function, generates 
#         Y = 2.1*X(0.16) - 0.2*X(0.47) - 1.9*X(0.67) + 5*X(0.85) + 4.2*X(0.91) + E
#   - fY3: Regression function log(1+t) with the standard L2 regression model
#
# Input
#   x: Data matrix, one row for each curve
#
# Return
#   y: Response vector
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

# Function trapz obtained fom the package pracma 
trapz <- function(x, y) {
  if (missing(y)) {
    if (length(x) == 0) return(0)
    y <- x
    x <- seq(along=x)
  }
  if (length(x) == 0 && length(y) == 0) return(0)
  if (!(is.numeric(x) || is.complex(x)) ||
      !(is.numeric(y) || is.complex(y)) )
    stop("Arguments 'x' and 'y' must be real or complex vectors.")
  m <- length(x)
  if (length(y) != m)
    stop("Arguments 'x', 'y' must be vectors of the same length.")
  if (m <= 1) return(0.0)
  
  xp <- c(x, x[m:1])
  yp <- c(numeric(m), y[m:1])
  n <- 2*m
  p1 <- sum(xp[1:(n-1)]*yp[2:n]) + xp[n]*yp[1]
  p2 <- sum(xp[2:n]*yp[1:(n-1)]) + xp[1]*yp[n]
  
  return(0.5*(p1-p2))
}


fY1 <- function(x) {
  return(c(2,-5,1)%*%x[c(20,40,90)])
}

fY2 <- function(x) {
  return(c(2.1, -0.2, -1.9, 5, 4.2)%*%x[c(16, 47, 60, 85, 91)])
}

fY3 <- function(x) {
  t = seq(0,1,length=length(x))
  funI = x*log(1+t)
  return(trapz(t,funI))
}


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
# generateBrown
#
# Generate a browninan motion data set and the associated responses using some of
# the previous regression functions
#
# Input
#   fY: Regression function
#   n (optional): Sample size, 200 by default
#   Nt (optional): Grid size, 100 by default
#   sig (optional): Variance of the marginal variables of the process, 1 by default
#   sigNoise (optional): Variance of the noise, 0.01 by default
#   geom (optional): If true a greometric Brownian Motion is generated, false by default
#
# Return
#   List of two elements:
#   - x: Data matrix, one row for each curve
#   - y: Vector of responses (real numbers)
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
generateBrown <- function(fY, n=200, Nt=100, sig=1, SigNoise=0.01, geom=FALSE) {
  x <- array(0, c(n,Nt))
  y <- rep(0, n)
  
  for (i in 1:n) {
    # Brownian motion starting in 1/Nt to avoid the degenerated point 0
    temp <- sqrt(1/ncol(x))*rnorm(n = Nt, sd = sqrt(sig))
    x[i,] <- cumsum(temp)
    if (geom) {x[i,] = exp(x[i,])}
    # Responses
    y[i] = fY(x[i,]) + rnorm(1,0,SigNoise)
  }
  
  return(list(x=x,y=y))
}


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
# generateIntegratedBrown
#
# Generate a integrated browninan motion data set and the associated responses using 
# some of the previous regression functions
#
# Input
#   fY: Regression function
#   n (optional): Sample size, 200 by default
#   Nt (optional): Grid size, 100 by default
#   sig (optional): Variance of the marginal variables of the process, 1 by default
#   sigNoise (optional): Variance of the noise, 0.01 by default
#
# Return
#   List of two elements:
#   - x: Data matrix, one row for each curve
#   - y: Vector of responses (real numbers)
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
generateIntegratedBrown <- function(fY, n=200, Nt=100, sig=1, SigNoise=0.01) {
  x <- array(0, c(n,Nt))
  y <- rep(0, n)
  
  for (i in 1:n) {
    # Brownian motion
    temp <- sqrt(1/(Nt+1))*rnorm(n = (Nt+1), sd = sqrt(sig)) #Para que empieze en t=1/Nt
    xtemp <- cumsum(temp)
    # Integrated brownian motion
    for (j in 1:Nt) {
      x[i,j] = trapz(1:(j+1),xtemp[1:(j+1)])
    }
    # Responses
    y[i] = fY(x[i,]) + rnorm(1,0,SigNoise)
  }
  
  return(list(x=x,y=y))
}


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
# generateOU
#
# Generate a Ornstein Uhlenbeck data set and the associated responses using some of
# the previous regression functions
#
# Input
#   fY: Regression function
#   n (optional): Sample size, 200 by default
#   Nt (optional): Grid size, 100 by default
#   mu (optional): Parameter of the OU process, 1 by default
#   theta (optional): Parameter of the OU process, 1 by default
#   sig (optional): Variance of the marginal variables of the process, 1 by default
#   sigNoise (optional): Variance of the noise, 0.01 by default
# 
# Return
#   List of two elements:
#   - x: Data matrix, one row for each curve
#   - y: Vector of responses (real numbers)
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

# Function HWV obtained fom the package Sim.DiffProc
HWV <- function(N, ...)  UseMethod("HWV")
HWV.default <- function(N =100,M=1,x0=2,t0=0,T=1,Dt,mu=4,theta=1,sigma=0.1,...)
{
  if (!is.numeric(x0)) stop("'x0' must be numeric")
  if (any(!is.numeric(t0) || !is.numeric(T))) stop(" 't0' and 'T' must be numeric")
  if (any(!is.numeric(N)  || (N - floor(N) > 0) || N <= 1)) stop(" 'N' must be a positive integer ")
  if (any(!is.numeric(M)  || (M - floor(M) > 0) || M <= 0)) stop(" 'M' must be a positive integer ")
  if (any(!is.numeric(sigma) || sigma <= 0) ) stop(" 'sigma' must be a numeric > 0 ")
  if (any(!is.numeric(mu) || mu <= 0) ) stop(" 'mu' must be a numeric > 0 ")
  if (any(t0 < 0 || T < 0 || T <= t0) ) stop(" please use positive times! (0 <= t0 < T) ")
  if (missing(Dt)) {
    t <- seq(t0, T, length = N + 1)
  } else {
    t <- c(t0, t0 + cumsum(rep(Dt, N)))
    T <- t[N + 1]
  }
  Dt <- (T - t0)/N
  hwv <- function()
  {
    w = c(0,cumsum(rnorm(N,mean=0,sd=sqrt(Dt))))
    dw   <- diff(w)
    X    <- numeric()
    X[1] <- x0
    Ito.sum <- c(0,sapply(1:(N+1),function(i){exp(-mu*Dt)*dw[i]}))
    X <- sapply(1:(N+1),function(i){theta+(x0-theta)*exp(-mu*t[i])+sigma*sum(Ito.sum[1:i])})
    X
  }
  res <- data.frame(sapply(1:M,function(i) hwv()))
  names(res) <- paste("X",1:M,sep="")
  X <- ts(res, start = t0, deltat = Dt)
  return(X)
}

# Function OU obtained fom the package Sim.DiffProc
OU <- function(N, ...)  UseMethod("OU")
OU.default <- function(N =100,M=1,x0=2,t0=0,T=1,Dt,mu=4,sigma=0.2,...)
{
  if (!is.numeric(x0)) stop("'x0' must be numeric")
  if (any(!is.numeric(t0) || !is.numeric(T))) stop(" 't0' and 'T' must be numeric")
  if (any(!is.numeric(N)  || (N - floor(N) > 0) || N <= 1)) stop(" 'N' must be a positive integer ")
  if (any(!is.numeric(M)  || (M - floor(M) > 0) || M <= 0)) stop(" 'M' must be a positive integer ")
  if (any(!is.numeric(sigma) || sigma <= 0) ) stop(" 'sigma' must be a numeric > 0 ")
  if (any(!is.numeric(mu) || mu <= 0) ) stop(" 'mu' must be a numeric > 0 ")
  if (any(t0 < 0 || T < 0 || T <= t0) ) stop(" please use positive times! (0 <= t0 < T) ")
  X <- HWV(N,M,x0,t0,T,Dt,mu,theta=0,sigma)
  return(X)
}


generateOU <- function(fY, n=200, Nt=100, mu=1, theta=1, sig=1, SigNoise=0.01) {
  library(sde)
  x <- array(0, c(n,Nt)) 
  y <- rep(0, n)
  # OU data matrix
  x = t(OU(N=Nt,M=n,mu=mu,sigma=sig,x0=0))[,2:(Nt+1)]
  # Responses
  for (i in 1:n) {
    y[i] = fY(x[i,]) + rnorm(1,0,SigNoise)
  }
  
  return(list(x=x,y=y))
}


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
# generateFBM
#
# Generate a Fractional Brownian Motion data set and the associated responses using 
# some of the previous regression functions
#
# Input
#   fY: Regression function
#   n (optional): Sample size, 200 by default
#   Nt (optional): Grid size, 100 by default
#   H (optional): Hurst index of the FBM, 0.5 by default 
#   sig (optional): Variance of the marginal variables of the process, 1 by default
#   sigNoise (optional): Variance of the noise, 0.01 by default
# 
# Return
#   List of two elements:
#   - x: Data matrix, one row for each curve
#   - y: Vector of responses (real numbers)
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

# Function circFBM obtained fom the package dvfBm
circFBM<- function(n, H, plotfBm=FALSE){
  if(missing(n)) n <- 500
  if(missing(H))
    H <- 0.6
  if(missing(plotfBm)) plotfBm <- 1       
  ## -------------------------------------------------------------------
  ## first line of the circulant matrix, C, built via covariances of fGn 
  ## -------------------------------------------------------------------
  lineC <- function(n, H, m){
    k <- 0:(m - 1)
    H2 <- 2 * H
    v <- (abs((k - 1)/n)^H2 - 2 * (k/n)^H2 + ((k + 1)/n)^H2)/2
    ind <- c(0:(m/2 - 1), m/2, (m/2 - 1):1)
    v <- v[ind + 1]
    drop(v)
  }
  ##
  ## ---------------------
  ## next power of two > n
  ## ---------------------
  m <- 2
  repeat {
    m <- 2 * m
    if(m >= (n - 1))
      break
  }
  stockm <- m     ##
  ## ----------------------------------------------
  ## research of the power of two (<2^18) such that
  ## C is definite positive
  ## ----------------------------------------------
  repeat {
    m <- 2 * m
    eigenvalC <- lineC(n, H, m)
    if (m<100) print(eigenvalC)
    eigenvalC <- Re(fft(c(eigenvalC), inverse = F))
    if((all(eigenvalC > 0)) | (m > 2^18))
      break
  }
  if(m > 2^18) {
    cat("----> exact method, impossible!!", fill = T)
    cat("----> can't find m such that C is definite positive", fill
        = T)
    break
  }
  else {
    ##
    ## -----------------------------------------------
    ## simulation of W=(Q)^t Z, where Z leads N(0,I_m)
    ## and  (Q)_{jk} = m^(-1/2) exp(-2i pi jk/m)
    ## -----------------------------------------------
    ar <- rnorm(m/2 + 1)
    ai <- rnorm(m/2 + 1)
    ar[1] <- sqrt(2) * ar[1]
    ar[(m/2 + 1)] <- sqrt(2) * ar[(m/2 + 1)]
    ai[1] <- 0
    ai[(m/2 + 1)] <- 0
    ar <- c(ar[c(1:(m/2 + 1))], ar[c((m/2):2)])
    aic <-  - ai
    ai <- c(ai[c(1:(m/2 + 1))], aic[c((m/2):2)])
    W <- complex(real = ar, imaginary = ai) ##
    ## -------------------------
    ## reconstruction of the fGn
    ## -------------------------
    W <- (sqrt(eigenvalC)) * W
    fGn <- fft(W, inverse = F)
    fGn <- (1/(sqrt(2 * m))) * fGn
    fGn <- Re(fGn[c(1:n)])
    fBm <- cumsum(fGn)
    fBm[1] <- 0     ##
    ## -----------
    ## plot of fBm
    ## -----------
    if(plotfBm) {
      par(mfrow = c(1, 1))
      time <- (0:(n - 1))/n
      Nchar <- as.character(n)
      Nleg <- paste(c("N= ", Nchar), collapse = " ")
      Hchar <- as.character(round(H, 3))
      Hleg <- paste(c(", H=", Hchar), collapse = "")
      NHleg <- paste(c(Nleg, Hleg), collapse = "")
      leg <- paste(c(
        "Path of a fractional Brownian motion ----- parameters",
        NHleg), collapse = " : ")
      plot(time, fBm, type = "l", main = leg)
    }
    fBm
  }
}


generateFBM <- function(fY,n=200,Nt=100,H=0.5,sig=1,SigNoise=0.01) {
  library(sde)
  x <- array(0, c(n,Nt)) 
  y <- rep(0, n)
  
  for (i in 1:n) {
    # FBM data
    x[i,] = circFBM((Nt+1),H,plotfBm=FALSE)[2:(Nt+1)] #Starting in 2 to avoid the 0
    # Response
    y[i] = fY(x[i,]) + rnorm(1,0,SigNoise)
  }
  
  return(list(x=x,y=y))
}


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
# readSugar
#
# Load the data concerning the ash content in sugar (section 6.2 of the paper)
#
# Input
#   center (optional): if true, the returned data has zero mean, true by default
#
# Return
#   List of two elements:
#   - x: Data matrix, one row for each curve
#   - y: Vector of responses (real numbers)
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
readSugar <- function(center=T) {
  library(R.matlab)
  sugar = readMat("sugar.mat")
  x = sugar$X[,100:600]
  y = sugar$y[,3]
  
  x = x[,5*(1:100)]
  
  # Remove the outliers, following Aneiros and Vieu (2014)
  x = x[c(1:70,72:nrow(x)),]
  y = y[c(1:70,72:length(y))] 
  
  n = nrow(x)
  Nt = ncol(x)
  
  # Remove the sample means if required
  if (center) {
    m = rep(NA, Nt)
    for (j in 1:Nt) {
      m[j] = mean(x[,j])
    }
    x = x - t(replicate(n,m))
    y = y - mean(y)
  }
  
  return(list(x=x,y=y))
}


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
# readTecator
#
# Load the centered second derivatives of the Tecator data set used in the peper 
# (section 6.2).
#
# Return
#   List of two elements:
#   - x: Data matrix, one row for each curve
#   - y: Vector of responses (real numbers)
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
readTecator <- function() {
  require(psych)
  data = read.file(f="tecator.txt", sep=" ")
  return(list(x=as.matrix(data[,1:100]),y=data[,101]))
}


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
# readTecator
#
# Load the current Tecator data set from de fda.usc package
#
# Input
#   center (optional): if true, the returned data has zero mean, true by default
#   derivative (optional): number of derivatives of the curves, 0 by default
#   remove.rep (optional): if true, the repited observations are removed, true 
#     by default
#
# Return
#   List of two elements:
#   - x: Data matrix, one row for each curve
#   - y: Vector of responses (real numbers)
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
readTecatorUSC <- function(center=T, derivative=0, remove.rep=T) {
  require(fda.usc)
  
  data(tecator)
  if (derivative<0) {
    stop("ERROR: only positive derivatives for the Tecator data set.")
  }
  
  if (derivative>0) {
    x = fdata.deriv(tecator$absorp.fdata,nderiv=derivative)$data
  } else {
    x = tecator$absorp.fdata$data
  }
  y = tecator$y$Fat
  
  n = nrow(x)
  Nt = ncol(x)
  
  # Remove the sample means if required
  if (center) {
    m = rep(NA, Nt)
    for (j in 1:Nt) {
      m[j] = mean(x[,j])
    }
    x = x - t(replicate(n,m))
    y = (y - mean(y))/sd(y)
  }
  
  # Remove the repeated observations if required
  if (remove.rep) {
    D = array(Inf,c(n,n))
    for (i in 1:(n-1)) {
      for (j in (i+1):n) {
        D[i,j] =  dist(rbind(x[i,],x[j,]))
      }
    }
    repited = which(D==0,arr.ind=T)
    x = x[setdiff(1:n,repited[,1]),]
    y = y[setdiff(1:n,repited[,1])]
  }
  
  return(list(x=x,y=y))
}


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
# readFlies
#
# Load the data containing the mediterranean fruit flies egg-laying profiles.
#
# Input
#   center (optional): if true, the returned data has zero mean, true by default
#
# Return
#   List of two elements:
#   - x: Data matrix, one row for each curve
#   - y: Vector of responses (real numbers)
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
readFlies <- function(center=T) {
  library(R.matlab)
  flies = readMat("MedFlies.mat")
  x = flies$X
  y = flies$Yreg
  
  n = nrow(x)
  Nt = ncol(x)
  
  # Remove the sample means if required
  if (center) {
    m = rep(NA, Nt)
    for (j in 1:Nt) {
      m[j] = mean(x[,j])
    }
    x = x - t(replicate(n,m))
    y = (y - mean(y))/sd(y)
  }
  
  return(list(x=x,y=y))
}