################################################################################
#################################################################################
# selectSeveralMethods.R
#
# Contains the function selectSeveralMethods
#
# Author: Beatriz Bueno-Larraz
# Date:   22-06-2016
#################################################################################
#################################################################################


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
# selectSeveralMethods
#
# Execute the variable selection methods exposed in Section 6.3 of the paper
#
# Input
#   x: Data matrix, one row for each curve.
#   y: Vector of response values.
#   params (optional): list containing the desired parameters for each of the methods. 
#     If some parameter is missing, the automatical values are used (described in each
#     of the method files). The elements of the list should be named as:
#     - RKHS: List of 3 elements containing Sigma, p and/or max
#     - KPS: List of 4 elements containing Sigma, p, max and/or kd
#     - PVS: List of 2 elements containing wn and/or max
#     - MH: List of 5 elements containing Sigma, p, max, H and/or criteria  
#   RKHS, KPS, PVS, MH (optional): Boolean value indicating if the method should be
#     executed or not. All the methods are executed by default.
#   mPar (optional): Only used when the methods are executed in parallel (several cores
#     at the same time). Indicate the number of the iteration in the parallelized loop.
#   time (optional): If true, the execution time of each method is measured.
#
# Return
#   List containing the following elements:
#     - RKHS, KPS, PVS, MH: Vectors of the selected points of each of the methods.
#       If some method is not executed, NULL is returned.
#     - tRKHS, tKPS, tPVS, tMH: Execution time if desired, NULL if not.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
selectSeveralMethods <- function(x, y, params=NULL, RKHS=T, KPS=T, PVS=T, MH=T, mPar=NULL, time=FALSE) {
  n = nrow(x)
  Nt = ncol(x)
  PRINT_ERROR = F #If true, some control lines are printed
  
  # Proposed method of the paper, using RKHS
  psRKHS=tRKHS=NULL
  if (RKHS) {
    if (PRINT_ERROR) print("Method RKHS")
    source("Methods/selRKHS.R")
    if (time) inicio = Sys.time()
    psRKHS <- (if (is.null(params$RKHS))  selRKHS(x,y)
             else if (!is.null(params$RKHS$Sigma) && is.null(params$RKHS$p) && is.null(params$RKHS$max)) 
               selRKHS(x,y,Sigma=params$RKHS$Sigma)
             else if (!is.null(params$RKHS$Sigma) && !is.null(params$RKHS$p)) 
               selRKHS(x,y,Sigma=params$RKHS$Sigma,p=params$RKHS$p)
             else if (!is.null(params$RKHS$Sigma) && is.null(params$RKHS$p) && !is.null(params$RKHS$max)) 
               selRKHS(x,y,Sigma=params$RKHS$Sigma,pMax=params$RKHS$max)
             else if (is.null(params$RKHS$Sigma) && !is.null(params$RKHS$p))
               selRKHS(x,y,p=params$RKHS$p)
             else if (is.null(params$RKHS$Sigma) && is.null(params$RKHS$p) && !is.null(params$RKHS$max))
               selRKHS(x,y,pMax=params$RKHS$max)
             else NULL)
    if (time) tRKHS = Sys.time() - inicio
  }
  
  # KPS method, proposed in Kneip et al. (2016) (Section 6.3 of the paper)
  psKPS = tKPS = NULL
  if (KPS) {
    if (PRINT_ERROR) print("Method KPS")
    source("Methods/selKPS.R")
    if (time) inicio = Sys.time()
    psKPS <- (if (is.null(params$KPS))  selKPS(x,y)
              else if (!is.null(params$KPS$Sigma) && is.null(params$KPS$max) && is.null(params$KPS$max)
                       && is.null(params$KPS$kd)) 
                selKPS(x,y,Sigma=params$KPS$Sigma)
              else if (!is.null(params$KPS$Sigma) && !is.null(params$KPS$p) && is.null(params$KPS$kd)) 
                selKPS(x,y,Sigma=params$KPS$Sigma,p=params$KPS$p)
              else if (!is.null(params$KPS$Sigma) && is.null(params$KPS$p) && !is.null(params$KPS$max)
                       && is.null(params$KPS$kd)) 
                selKPS(x,y,Sigma=params$KPS$Sigma,pMax=params$KPS$max)
              else if (is.null(params$KPS$Sigma) && !is.null(params$KPS$p) && is.null(params$KPS$kd))
                selKPS(x,y,p=params$KPS$p)
              else if (is.null(params$KPS$Sigma) && is.null(params$KPS$p) && !is.null(params$KPS$max)
                       && is.null(params$KPS$kd))
                selKPS(x,y,pMax=params$KPS$max)
              else if (!is.null(params$KPS$Sigma) && is.null(params$KPS$p) && is.null(params$KPS$max)
                       && !is.null(params$KPS$kd)) 
                selKPS(x,y,Sigma=params$KPS$Sigma)
              else if (!is.null(params$KPS$Sigma) && !is.null(params$KPS$p) && !is.null(params$KPS$kd)) 
                selKPS(x,y,Sigma=params$KPS$Sigma,p=params$KPS$p)
              else if (!is.null(params$KPS$Sigma) && is.null(params$KPS$p) && !is.null(params$KPS$max)
                       && !is.null(params$KPS$kd)) 
                selKPS(x,y,Sigma=params$KPS$Sigma,pMax=params$KPS$max)
              else if (is.null(params$KPS$Sigma) && !is.null(params$KPS$p) && !is.null(params$KPS$kd))
                selKPS(x,y,p=params$KPS$p)
              else if (is.null(params$KPS$Sigma) && is.null(params$KPS$p) && !is.null(params$KPS$max)
                       && !is.null(params$KPS$kd))
                selKPS(x,y,pMax=params$KPS$max)
              else NULL)
    if (time) tKPS = Sys.time() - inicio
  }
  
  
  # PVS method, proposed in Aneiros and Vieu (2014) (Section 6.3 of the paper)
  psPVS = tPVS = NULL
  if (PVS) {
    if (PRINT_ERROR) print("Method PVS")
    source("Methods/selPVS.R")
    if (time) inicio = Sys.time()
    psPVS <- (if (is.null(params$PVS))  selPVS(x,y)
            else if (!is.null(params$PVS$wn))
              selPVS(x,y,wn=params$PVS$wn)
            else if (!is.null(params$PVS$max))
              selPVS(x,y,wnMax=params$PVS$max)
            else NULL)
    if (time) tPVS = Sys.time() - inicio
  }
  
  
  # MH method, proposed in Berrendero et al. (2016) (Section 6.3 of the paper)
  psMH = tMH = NULL
  if (MH) {
    if (is.null(mPar)) {
      if (PRINT_ERROR) print("Method MH")
      source("Methods/selMH.R")
      if (time) inicio = Sys.time()
      psMH <- (if (is.null(params$MH))  selMH(x,y)
               else if (!is.null(params$MH$p) && !is.null(params$MH$Sigma) && is.null(params$MH$H) && is.null(params$MH$criteria)) 
                 selMH(x,y,p=params$MH$p,Sigma=params$MH$Sigma)
               else if (!is.null(params$MH$p) && !is.null(params$MH$Sigma) && is.null(params$MH$H) && !is.null(params$MH$criteria)) 
                 selMH(x,y,p=params$MH$p,Sigma=params$MH$Sigma,criteria=params$MH$criteria)
               else if (!is.null(params$MH$p) && !is.null(params$MH$Sigma) && !is.null(params$MH$H) && is.null(params$MH$criteria)) 
                 selMH(x,y,p=params$MH$p,Sigma=params$MH$Sigma,H=params$MH$H)
               else if (!is.null(params$MH$p) && !is.null(params$MH$Sigma) && !is.null(params$MH$H) && !is.null(params$MH$criteria)) 
                 selMH(x,y,p=params$MH$p,Sigma=params$MH$Sigma,H=params$MH$H,criteria=params$MH$criteria)
               else if (!is.null(params$MH$p) && is.null(params$MH$Sigma) && is.null(params$MH$H) && !is.null(params$MH$criteria))
                 selMH(x,y,p=params$MH$p)
               else if (!is.null(params$MH$p) && is.null(params$MH$Sigma) && is.null(params$MH$H) && !is.null(params$MH$criteria)) 
                 selMH(x,y,p=params$MH$p,criteria=params$MH$criteria)
               else if (!is.null(params$MH$p) && is.null(params$MH$Sigma) && !is.null(params$MH$H) && is.null(params$MH$criteria)) 
                 selMH(x,y,p=params$MH$p,H=params$MH$H)
               else if (!is.null(params$MH$p) && is.null(params$MH$Sigma) && !is.null(params$MH$H) && !is.null(params$MH$criteria)) 
                 selMH(x,y,p=params$MH$p,H=params$MH$H,criteria=params$MH$criteria)
               else if (is.null(params$MH$p) && !is.null(params$MH$Sigma) && is.null(params$MH$max) && is.null(params$MH$H) && is.null(params$MH$criteria)) 
                 selMH(x,y,Sigma=params$MH$Sigma)
               else if (is.null(params$MH$p) && !is.null(params$MH$Sigma) && is.null(params$MH$max) && is.null(params$MH$H) && !is.null(params$MH$criteria)) 
                 selMH(x,y,Sigma=params$MH$Sigma,criteria=params$MH$criteria)
               else if (is.null(params$MH$p) && !is.null(params$MH$Sigma) && is.null(params$MH$max) && !is.null(params$MH$H) && is.null(params$MH$criteria)) 
                 selMH(x,y,Sigma=params$MH$Sigma,H=params$MH$H)
               else if (is.null(params$MH$p) && !is.null(params$MH$Sigma) && is.null(params$MH$max) && !is.null(params$MH$H) && !is.null(params$MH$criteria)) 
                 selMH(x,y,Sigma=params$MH$Sigma,H=params$MH$H,criteria=params$MH$criteria)
               else if (is.null(params$MH$p) && !is.null(params$MH$Sigma) && !is.null(params$MH$max)  && is.null(params$MH$H) && is.null(params$MH$criteria)) 
                 selMH(x,y,Sigma=params$MH$Sigma,pMax=params$MH$max)
               else if (is.null(params$MH$p) && !is.null(params$MH$Sigma) && !is.null(params$MH$max)  && is.null(params$MH$H) && !is.null(params$MH$criteria)) 
                 selMH(x,y,Sigma=params$MH$Sigma,pMax=params$MH$max,criteria=params$MH$criteria)
               else if (is.null(params$MH$p) && !is.null(params$MH$Sigma) && !is.null(params$MH$max)  && !is.null(params$MH$H) && is.null(params$MH$criteria)) 
                 selMH(x,y,Sigma=params$MH$Sigma,pMax=params$MH$max,H=params$MH$H)
               else if (is.null(params$MH$p) && !is.null(params$MH$Sigma) && !is.null(params$MH$max)  && !is.null(params$MH$H) && !is.null(params$MH$criteria)) 
                 selMH(x,y,Sigma=params$MH$Sigma,pMax=params$MH$max,H=params$MH$H,criteria=params$MH$criteria)
               else if (is.null(params$MH$p) && is.null(params$MH$Sigma) && is.null(params$MH$max) && is.null(params$MH$H) && !is.null(params$MH$criteria)) 
                 selMH(x,y,criteria=params$MH$criteria)
               else if (is.null(params$MH$p) && is.null(params$MH$Sigma) && is.null(params$MH$max) && !is.null(params$MH$H) && is.null(params$MH$criteria)) 
                 selMH(x,y,H=params$MH$H)
               else if (is.null(params$MH$p) && is.null(params$MH$Sigma) && is.null(params$MH$max) && !is.null(params$MH$H) && !is.null(params$MH$criteria)) 
                 selMH(x,y,H=params$MH$H,criteria=params$MH$criteria)
               else if (is.null(params$MH$p) && is.null(params$MH$Sigma) && !is.null(params$MH$max)  && is.null(params$MH$H) && is.null(params$MH$criteria)) 
                 selMH(x,y,pMax=params$MH$max)
               else if (is.null(params$MH$p) && is.null(params$MH$Sigma) && !is.null(params$MH$max)  && is.null(params$MH$H) && !is.null(params$MH$criteria)) 
                 selMH(x,y,pMax=params$MH$max,criteria=params$MH$criteria)
               else if (is.null(params$MH$p) && is.null(params$MH$Sigma) && !is.null(params$MH$max)  && !is.null(params$MH$H) && is.null(params$MH$criteria)) 
                 selMH(x,y,pMax=params$MH$max,H=params$MH$H)
               else if (is.null(params$MH$p) && is.null(params$MH$Sigma) && !is.null(params$MH$max)  && !is.null(params$MH$H) && !is.null(params$MH$criteria)) 
                 selMH(x,y,pMax=params$MH$max,H=params$MH$H,criteria=params$MH$criteria)
               else NULL)
    } else {
      if (PRINT_ERROR) print("Method MH")
      source("Methods/selMH_Par.R")
      if (time) inicio = Sys.time()
      psMH <- (if (is.null(params$MH))  selMH_Par(mPar=mPar,x,y)
               else if (!is.null(params$MH$p) && !is.null(params$MH$Sigma) && is.null(params$MH$H) && is.null(params$MH$criteria)) 
                 selMH_Par(mPar=mPar,x,y,p=params$MH$p,Sigma=params$MH$Sigma)
               else if (!is.null(params$MH$p) && !is.null(params$MH$Sigma) && is.null(params$MH$H) && !is.null(params$MH$criteria)) 
                 selMH_Par(mPar=mPar,x,y,p=params$MH$p,Sigma=params$MH$Sigma,criteria=params$MH$criteria)
               else if (!is.null(params$MH$p) && !is.null(params$MH$Sigma) && !is.null(params$MH$H) && is.null(params$MH$criteria)) 
                 selMH_Par(mPar=mPar,x,y,p=params$MH$p,Sigma=params$MH$Sigma,H=params$MH$H)
               else if (!is.null(params$MH$p) && !is.null(params$MH$Sigma) && !is.null(params$MH$H) && !is.null(params$MH$criteria)) 
                 selMH_Par(mPar=mPar,x,y,p=params$MH$p,Sigma=params$MH$Sigma,H=params$MH$H,criteria=params$MH$criteria)
               else if (!is.null(params$MH$p) && is.null(params$MH$Sigma) && is.null(params$MH$H) && !is.null(params$MH$criteria))
                 selMH_Par(mPar=mPar,x,y,p=params$MH$p)
               else if (!is.null(params$MH$p) && is.null(params$MH$Sigma) && is.null(params$MH$H) && !is.null(params$MH$criteria)) 
                 selMH_Par(mPar=mPar,x,y,p=params$MH$p,criteria=params$MH$criteria)
               else if (!is.null(params$MH$p) && is.null(params$MH$Sigma) && !is.null(params$MH$H) && is.null(params$MH$criteria)) 
                 selMH_Par(mPar=mPar,x,y,p=params$MH$p,H=params$MH$H)
               else if (!is.null(params$MH$p) && is.null(params$MH$Sigma) && !is.null(params$MH$H) && !is.null(params$MH$criteria)) 
                 selMH_Par(mPar=mPar,x,y,p=params$MH$p,H=params$MH$H,criteria=params$MH$criteria)
               else if (is.null(params$MH$p) && !is.null(params$MH$Sigma) && is.null(params$MH$max) && is.null(params$MH$H) && is.null(params$MH$criteria)) 
                 selMH_Par(mPar=mPar,x,y,Sigma=params$MH$Sigma)
               else if (is.null(params$MH$p) && !is.null(params$MH$Sigma) && is.null(params$MH$max) && is.null(params$MH$H) && !is.null(params$MH$criteria)) 
                 selMH_Par(mPar=mPar,x,y,Sigma=params$MH$Sigma,criteria=params$MH$criteria)
               else if (is.null(params$MH$p) && !is.null(params$MH$Sigma) && is.null(params$MH$max) && !is.null(params$MH$H) && is.null(params$MH$criteria)) 
                 selMH_Par(mPar=mPar,x,y,Sigma=params$MH$Sigma,H=params$MH$H)
               else if (is.null(params$MH$p) && !is.null(params$MH$Sigma) && is.null(params$MH$max) && !is.null(params$MH$H) && !is.null(params$MH$criteria)) 
                 selMH_Par(mPar=mPar,x,y,Sigma=params$MH$Sigma,H=params$MH$H,criteria=params$MH$criteria)
               else if (is.null(params$MH$p) && !is.null(params$MH$Sigma) && !is.null(params$MH$max)  && is.null(params$MH$H) && is.null(params$MH$criteria)) 
                 selMH_Par(mPar=mPar,x,y,Sigma=params$MH$Sigma,pMax=params$MH$max)
               else if (is.null(params$MH$p) && !is.null(params$MH$Sigma) && !is.null(params$MH$max)  && is.null(params$MH$H) && !is.null(params$MH$criteria)) 
                 selMH_Par(mPar=mPar,x,y,Sigma=params$MH$Sigma,pMax=params$MH$max,criteria=params$MH$criteria)
               else if (is.null(params$MH$p) && !is.null(params$MH$Sigma) && !is.null(params$MH$max)  && !is.null(params$MH$H) && is.null(params$MH$criteria)) 
                 selMH_Par(mPar=mPar,x,y,Sigma=params$MH$Sigma,pMax=params$MH$max,H=params$MH$H)
               else if (is.null(params$MH$p) && !is.null(params$MH$Sigma) && !is.null(params$MH$max)  && !is.null(params$MH$H) && !is.null(params$MH$criteria)) 
                 selMH_Par(mPar=mPar,x,y,Sigma=params$MH$Sigma,pMax=params$MH$max,H=params$MH$H,criteria=params$MH$criteria)
               else if (is.null(params$MH$p) && is.null(params$MH$Sigma) && is.null(params$MH$max) && is.null(params$MH$H) && !is.null(params$MH$criteria)) 
                 selMH_Par(mPar=mPar,x,y,criteria=params$MH$criteria)
               else if (is.null(params$MH$p) && is.null(params$MH$Sigma) && is.null(params$MH$max) && !is.null(params$MH$H) && is.null(params$MH$criteria)) 
                 selMH_Par(mPar=mPar,x,y,H=params$MH$H)
               else if (is.null(params$MH$p) && is.null(params$MH$Sigma) && is.null(params$MH$max) && !is.null(params$MH$H) && !is.null(params$MH$criteria)) 
                 selMH_Par(mPar=mPar,x,y,H=params$MH$H,criteria=params$MH$criteria)
               else if (is.null(params$MH$p) && is.null(params$MH$Sigma) && !is.null(params$MH$max)  && is.null(params$MH$H) && is.null(params$MH$criteria)) 
                 selMH_Par(mPar=mPar,x,y,pMax=params$MH$max)
               else if (is.null(params$MH$p) && is.null(params$MH$Sigma) && !is.null(params$MH$max)  && is.null(params$MH$H) && !is.null(params$MH$criteria)) 
                 selMH_Par(mPar=mPar,x,y,pMax=params$MH$max,criteria=params$MH$criteria)
               else if (is.null(params$MH$p) && is.null(params$MH$Sigma) && !is.null(params$MH$max)  && !is.null(params$MH$H) && is.null(params$MH$criteria)) 
                 selMH_Par(mPar=mPar,x,y,pMax=params$MH$max,H=params$MH$H)
               else if (is.null(params$MH$p) && is.null(params$MH$Sigma) && !is.null(params$MH$max)  && !is.null(params$MH$H) && !is.null(params$MH$criteria)) 
                 selMH_Par(mPar=mPar,x,y,pMax=params$MH$max,H=params$MH$H,criteria=params$MH$criteria)
               else NULL)
    }
    if (time) tMH = Sys.time() - inicio
  }
  
  
  return(list(RKHS=psRKHS, KPS=psKPS, PVS=psPVS, MH=psMH,
              tRKHS=tRKHS, tKPS=tKPS, tPVS=tPVS, tMH=tMH))
}